"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = simplifyAbstractValue;

var _errors = require("../errors.js");

var _index = require("../domains/index.js");

var _invariant = require("../invariant.js");

var _invariant2 = _interopRequireDefault(_invariant);

var _realm = require("../realm.js");

var _index2 = require("../values/index.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function simplifyAbstractValue(realm, value) {
  var savedHandler = realm.errorHandler;
  var savedIsReadOnly = realm.isReadOnly;
  realm.isReadOnly = true;
  try {
    realm.errorHandler = function () {
      throw new _errors.FatalError();
    };
    return simplify(realm, value);
  } catch (e) {
    return value;
  } finally {
    realm.errorHandler = savedHandler;
    realm.isReadOnly = savedIsReadOnly;
  }
} /**
   * Copyright (c) 2017-present, Facebook, Inc.
   * All rights reserved.
   *
   * This source code is licensed under the BSD-style license found in the
   * LICENSE file in the root directory of this source tree. An additional grant
   * of patent rights can be found in the PATENTS file in the same directory.
   */

function simplify(realm, value) {
  if (value instanceof _index2.ConcreteValue) return value;
  (0, _invariant2.default)(value instanceof _index2.AbstractValue);
  switch (value.kind) {
    case "!":
      return negate(realm, value.args[0]);
    default:
      return value;
  }
}

function negate(realm, value) {
  if (value instanceof _index2.ConcreteValue) return _index.ValuesDomain.computeUnary(realm, "!", value);
  (0, _invariant2.default)(value instanceof _index2.AbstractValue);
  if (value.kind === "!") return value.args[0];
  // todo: remove this check once intrinsic objects can be properly nullable
  if (!(value instanceof _index2.AbstractObjectValue) || !value.isIntrinsic()) {
    if (!value.mightNotBeTrue()) return realm.intrinsics.false;
    if (!value.mightNotBeFalse()) return realm.intrinsics.true;
  }
  // If NaN is not an issue, invert binary ops
  if (value.args.length === 2 && !value.args[0].mightBeNumber() && !value.args[1].mightBeNumber()) {
    var invertedComparison = void 0;
    switch (value.kind) {
      case "===":
        invertedComparison = "!==";
        break;
      case "==":
        invertedComparison = "!=";
        break;
      case "!==":
        invertedComparison = "===";
        break;
      case "!=":
        invertedComparison = "==";
        break;
      case "<":
        invertedComparison = ">=";
        break;
      case "<=":
        invertedComparison = ">";
        break;
      case ">":
        invertedComparison = "<=";
        break;
      case ">=":
        invertedComparison = "<";
        break;
      default:
        break;
    }
    if (invertedComparison !== undefined) {
      var left = simplify(realm, value.args[0]);
      var right = simplify(realm, value.args[1]);
      return _index2.AbstractValue.createFromBinaryOp(realm, invertedComparison, left, right, value.expressionLocation);
    }
    var invertedLogicalOp = void 0;
    switch (value.kind) {
      case "&&":
        invertedLogicalOp = "||";
        break;
      case "||":
        invertedLogicalOp = "&&";
        break;
      default:
        break;
    }
    if (invertedLogicalOp !== undefined) {
      var _left = negate(realm, value.args[0]);
      var _right = negate(realm, value.args[1]);
      return _index2.AbstractValue.createFromLogicalOp(realm, invertedLogicalOp, _left, _right, value.expressionLocation);
    }
  }
  return _index2.AbstractValue.createFromUnaryOp(realm, "!", value, true, value.expressionLocation);
}
//# sourceMappingURL=simplifier.js.map